from __future__ import absolute_import, print_function, unicode_literals

import functools
from django.apps import apps
from django.utils import module_loading, six


def autodiscover(*args):
    module_loading.autodiscover_modules('handlers', *args)


class Handler(object):

    _handler = None

    def __init__(self, func):
        functools.wraps(func)(self)
        self._handler = func

    def __call__(self, *args, **kwargs):
        return self._handler(*args, **kwargs)

    def connect(self, signal, **kwargs):
        sender = kwargs.pop('sender', None)
        if isinstance(sender, six.string_types):
            sender = apps.get_model(sender)
        signal.connect(self._handler, sender=sender, **kwargs)

    def disconnect(self, signal, **kwargs):
        # https://docs.djangoproject.com/en/2.0/topics/signals/#disconnecting-signals
        sender = kwargs.pop('sender', None)
        if isinstance(sender, six.string_types):
            sender = apps.get_model(sender)
        return signal.disconnect(self._handler, sender=sender, **kwargs)


def handler(*args, **kwargs):
    if len(args) == 1 and callable(args[0]):
        return Handler(args[0])

    def decorator(func):
        wrapped = Handler(func)
        wrapped.connect(*args, **kwargs)
        return wrapped

    return decorator



from __future__ import absolute_import, print_function, unicode_literals

from django.db.models.signals import post_init, post_save
from django.dispatch import Signal
from django.test import TestCase
from tests.myapp.models import Foo, test_handler
from roadies.handlers import handler, Handler

try:
    import mock
except ImportError:
    from unittest import mock


def async_action(*args):
    pass


@handler
def do_something(sender, foo, bar, **kwargs):
    """Test functional handlers."""
    async_action(foo, bar)


@handler(post_init, sender='myapp.Foo', dispatch_uid='testing')
def do_something_else(sender, instance, **kwargs):
    async_action(instance)


test_signal = Signal(providing_args=['foo', 'bar'])


class HandlerTests(TestCase):

    def test_do_something_is_handler_instance(self):
        self.assertIsInstance(do_something, Handler)

    def test_connect_no_sender(self):
        do_something.connect(test_signal, dispatch_uid='testing')
        result = do_something.disconnect(test_signal, dispatch_uid='testing')
        self.assertTrue(result)
        result = do_something.disconnect(test_signal, dispatch_uid='testing')
        self.assertFalse(result)

    def test_connect_with_sender_string(self):
        do_something.connect(
            test_signal,
            sender='myapp.Foo',
            dispatch_uid='testing'
        )
        disconnected = do_something.disconnect(
            test_signal,
            sender='myapp.Foo',
            dispatch_uid='testing'
        )
        self.assertTrue(disconnected)

        disconnected = do_something.disconnect(
            test_signal,
            sender='myapp.Foo',
            dispatch_uid='testing'
        )
        self.assertFalse(disconnected)

    def test_do_something_name(self):
        self.assertEqual(do_something.__name__, 'do_something')

    def test_do_something_module(self):
        self.assertEqual(do_something.__module__, 'tests.test_handlers')

    def test_do_something_doc(self):
        self.assertEqual(do_something.__doc__, 'Test functional handlers.')

    def test_do_something_does_something_on_signal(self):
        do_something.connect(test_signal, dispatch_uid='testing')
        with mock.patch('tests.test_handlers.async_action') as func:
            test_signal.send(None, foo=1, bar=2)
            self.assertTrue(func.called)
            func.assert_called_once_with(1, 2)

    def test_do_something_else_is_handler(self):
        self.assertIsInstance(do_something_else, Handler)

    def test_do_something_else_executed_on_post_init(self):
        with mock.patch('tests.test_handlers.async_action') as func:
            instance = Foo(bar='baz')
            self.assertTrue(func.called)
            func.assert_called_once_with(instance)

    def test_calling_the_handler_directly(self):
        with mock.patch('tests.test_handlers.async_action') as func:
            do_something(None, 'foo', 'bar')
            self.assertTrue(func.called)
            func.assert_called_once_with('foo', 'bar')

    def test_autodiscover_works(self):
        self.assertTrue(
            post_save.disconnect(test_handler, dispatch_uid='test_handler')
        )



from __future__ import absolute_import, print_function, unicode_literals

from django.db import models


class Foo(models.Model):

    bar = models.CharField(max_length=16)

    def __str__(self):
        return self.bar


def test_handler(sender, **kwargs):
    pass



from __future__ import absolute_import, print_function, unicode_literals

from django.db.utils import OperationalError
from roadies.test import TempModel, TestCase

from .models import Person, Place, ThingWithName


class TestTempModels(TestCase):

    TEST_MODELS = (Person, Place, Person)

    def test_model_inheritance(self):
        self.assertIsInstance(Person(), ThingWithName)
        self.assertIsInstance(Person(), TempModel)

    def test_model_save(self):
        person = Person(name='bob')
        self.assertIsNone(person.pk)
        person.save()
        self.assertIsNotNone(person.pk)


class TestTempModelsRepeated(TestCase):

    TEST_MODELS = (Place, )

    def test_place_save(self):
        place = Place(name='bob', latitude=0, longitude=0)
        self.assertIsNone(place.pk)
        place.save()
        self.assertIsNotNone(place.pk)

    def test_person_save_raises(self):
        # Person is not defined within the TestCase and is therefore out of scope
        person = Person(name='bob')
        self.assertRaises(OperationalError, person.save)


class TestBackwardsCompatibility(TestCase):

    def test_success(self):
        self.assertTrue(True)



from __future__ import absolute_import, print_function, unicode_literals

from django.db.models.signals import post_save
from .models import test_handler

post_save.connect(test_handler, dispatch_uid='test_handler')



from __future__ import absolute_import, print_function, unicode_literals


from django.test import TestCase
from roadies.test import mock, OnCommitMock

from .models import do_it_later


class TestOnCommitMock(TestCase):

    def test_self_is_returned_on_enter(self):
        with OnCommitMock('tests.models.transaction') as on_commit:
            self.assertIsInstance(on_commit, OnCommitMock)

    def test_add_a_function(self):
        func = mock.MagicMock()
        with OnCommitMock('tests.models.transaction') as on_commit:
            do_it_later(func)
            self.assertTrue(func in on_commit.functions)

    def test_function_called_on_exit(self):
        func = mock.MagicMock()
        with OnCommitMock('tests.models.transaction'):
            do_it_later(func)
            self.assertFalse(func.called)
        self.assertTrue(func.called)

    def test_trigger_idempotence(self):
        func = mock.MagicMock()
        with OnCommitMock('tests.models.transaction') as on_commit:
            do_it_later(func)
            self.assertFalse(func.called)
            on_commit.trigger()
            self.assertTrue(func.called)
            self.assertEqual(func.call_count, 1)
            on_commit.trigger()
            self.assertEqual(func.call_count, 1)

    def test_raises_on_incorrect_target(self):
        self.assertRaises(TypeError, OnCommitMock, 1)



from __future__ import absolute_import, print_function, unicode_literals

from django.apps import AppConfig
from roadies import handlers


class MyAppConfig(AppConfig):

    name = 'tests.myapp'

    def ready(self):
        handlers.autodiscover()
